import operator
from datetime import datetime
from functools import reduce
from typing import Self

from libecalc.application.graph_result import GraphResult
from libecalc.common.time_utils import Frequency, Period, Periods
from libecalc.common.units import Unit
from libecalc.common.utils.rates import TimeSeriesFloat, TimeSeriesStreamDayRate
from libecalc.common.variables import ExpressionEvaluator, VariablesMap
from libecalc.core.result import EcalcModelResult
from libecalc.core.result.emission import EmissionResult
from libecalc.domain.component_validation_error import DomainValidationException
from libecalc.domain.energy import ComponentEnergyContext, Emitter, EnergyComponent, EnergyModel
from libecalc.dto import ResultOptions
from libecalc.dto.component_graph import ComponentGraph
from libecalc.presentation.yaml.domain.reference_service import ReferenceService
from libecalc.presentation.yaml.domain.time_series_collections import TimeSeriesCollections
from libecalc.presentation.yaml.mappers.component_mapper import EcalcModelMapper
from libecalc.presentation.yaml.mappers.create_references import create_references
from libecalc.presentation.yaml.mappers.variables_mapper import map_yaml_to_variables
from libecalc.presentation.yaml.mappers.variables_mapper.get_global_time_vector import get_global_time_vector
from libecalc.presentation.yaml.mappers.variables_mapper.variables_mapper import InvalidVariablesException
from libecalc.presentation.yaml.model_validation_exception import ModelValidationException
from libecalc.presentation.yaml.resource_service import ResourceService
from libecalc.presentation.yaml.validation_errors import (
    DataValidationError,
    DtoValidationError,
    Location,
    ModelValidationError,
)
from libecalc.presentation.yaml.yaml_keywords import EcalcYamlKeywords
from libecalc.presentation.yaml.yaml_models.yaml_model import YamlValidator
from libecalc.presentation.yaml.yaml_validation_context import (
    ModelContext,
    ModelName,
    YamlModelValidationContext,
    YamlModelValidationContextNames,
)

DEFAULT_START_TIME = datetime(1900, 1, 1)


class Context(ComponentEnergyContext):
    def __init__(
        self,
        energy_model: EnergyModel,
        consumer_results: dict[str, EcalcModelResult],
        component_id: str,
    ):
        self._energy_model = energy_model
        self._consumer_results = consumer_results
        self._component_id = component_id

    def get_power_requirement(self) -> TimeSeriesFloat | None:
        consumer_power_usage = [
            self._consumer_results[consumer.id].component_result.power
            for consumer in self._energy_model.get_consumers(self._component_id)
            if self._consumer_results[consumer.id].component_result.power is not None
        ]

        if len(consumer_power_usage) < 1:
            return None

        if len(consumer_power_usage) == 1:
            return consumer_power_usage[0]

        return reduce(operator.add, consumer_power_usage)

    def get_fuel_usage(self) -> TimeSeriesStreamDayRate | None:
        energy_usage = self._consumer_results[self._component_id].component_result.energy_usage
        if energy_usage.unit == Unit.MEGA_WATT:
            # energy usage is power usage, not fuel usage.
            return None
        return energy_usage


class YamlModel(EnergyModel):
    """
    Class representing both the yaml and the resources.

    We haven't defined a difference in naming between the YamlModel representing only the yaml file and this class,
    which also have information about the referenced resources.

    Maybe we could use 'configuration' for the single yaml file, that naming is already used a lot for the instantiation
    of that YamlModel class.

    configuration: the model configuration
    resources: the model 'input', kind of
    model: configuration + resources (input)
    """

    def __init__(
        self,
        configuration: YamlValidator,
        resource_service: ResourceService,
        output_frequency: Frequency,
    ) -> None:
        self._output_frequency = output_frequency
        self._configuration = configuration
        self.resources = resource_service.get_resources(self._configuration)

        self._is_validated = False
        self._graph = None
        self._consumer_results: dict[str, EcalcModelResult] = {}
        self._emission_results: dict[str, dict[str, EmissionResult]] = {}

    def get_consumers(self, provider_id: str = None) -> list[EnergyComponent]:
        return self._get_graph().get_consumers(provider_id)

    def get_energy_components(self) -> list[EnergyComponent]:
        return self._get_graph().get_energy_components()

    def get_expression_evaluator(self) -> ExpressionEvaluator:
        return self.variables

    def _get_reference_service(self) -> ReferenceService:
        return create_references(self._configuration, self.resources)

    @property
    def period(self) -> Period:
        return Period(
            start=self.start or DEFAULT_START_TIME,
            end=self.end or datetime.max,
        )

    @property
    def start(self) -> datetime | None:
        return self._configuration.start

    @property
    def end(self) -> datetime | None:
        return self._configuration.end

    def _get_time_series_collections(self) -> TimeSeriesCollections:
        time_series_collections, err = TimeSeriesCollections.create(
            time_series=self._configuration.time_series,
            resources=self.resources,
            raise_on_error=True,
        )
        assert len(err) == 0
        return time_series_collections

    def _get_periods(self) -> Periods:
        time_vector = get_global_time_vector(
            time_series_time_vector=self._get_time_series_collections().get_time_vector(),
            start=self._configuration.start,
            end=self._configuration.end,
            frequency=self._output_frequency,
            additional_dates=self._configuration.dates,
        )
        return Periods.create_periods(time_vector, include_before=False, include_after=False)

    @property
    def variables(self) -> VariablesMap:
        return map_yaml_to_variables(
            configuration=self._configuration,
            time_series_provider=self._get_time_series_collections(),
            periods=self._get_periods(),
        )

    @property
    def result_options(self) -> ResultOptions:
        return ResultOptions(
            start=self._configuration.start,
            end=self._configuration.end,
            output_frequency=self._output_frequency,
        )

    def _get_graph(self) -> ComponentGraph:
        if self._is_validated and self._graph is not None:
            return self._graph

        self.validate_for_run()
        model_mapper = EcalcModelMapper(
            references=self._get_reference_service(),
            target_period=self.period,
            expression_evaluator=self.variables,
        )

        dto = model_mapper.from_yaml_to_domain(configuration=self._configuration)
        self._graph = dto.get_graph()
        return self._graph

    def _get_token_references(self, yaml_model: YamlValidator) -> list[str]:
        # Only get references for valid time series collections
        time_series_collections, _ = TimeSeriesCollections.create(
            time_series=self._configuration.time_series, resources=self.resources, raise_on_error=False
        )
        token_references = time_series_collections.get_time_series_references()

        for reference in yaml_model.variables:
            token_references.append(f"$var.{reference}")

        return token_references

    @staticmethod
    def _get_model_types(yaml_model: YamlValidator) -> dict["ModelName", "ModelContext"]:
        models = [*yaml_model.models, *yaml_model.facility_inputs]
        model_types: dict[ModelName, ModelContext] = {}
        for model in models:
            if hasattr(model, "name"):
                model_types[model.name] = model
        return model_types

    def _get_validation_context(self, yaml_model: YamlValidator) -> YamlModelValidationContext:
        return {
            YamlModelValidationContextNames.model_name: yaml_model.name,  # type: ignore[misc]
            YamlModelValidationContextNames.resource_file_names: [name for name, resource in self.resources.items()],
            YamlModelValidationContextNames.expression_tokens: self._get_token_references(yaml_model=yaml_model),
            YamlModelValidationContextNames.model_types: self._get_model_types(yaml_model=yaml_model),
        }

    def validate_for_run(self) -> Self:
        if self._is_validated:
            return self

        try:
            # Validate model
            validation_context = self._get_validation_context(yaml_model=self._configuration)
            self._configuration.validate(validation_context)

            # Is validated, first step, must be set before calling get_graph
            self._is_validated = True

            # Validate and create the graph used for evaluating the energy model
            self._get_graph()
            return self
        except InvalidVariablesException as e:
            # TODO: Variables are evaluated when setting up ExpressionEvaluator. This seems unnecessary.
            #  We could evaluate when needed instead.
            raise ModelValidationException(
                errors=[
                    ModelValidationError(
                        location=Location(keys=[EcalcYamlKeywords.variables]),
                        message=str(e),
                        data=None,
                        file_context=None,
                    )
                ],
            ) from e
        except (
            DtoValidationError,
            DomainValidationException,
        ) as e:
            raise ModelValidationException(errors=e.errors()) from e  # type: ignore[arg-type]
        except DataValidationError as e:
            raise ModelValidationException(
                errors=[
                    ModelValidationError(
                        location=Location(keys=[""]),
                        message=str(e),
                        data=None,
                        file_context=None,
                    )
                ],
            ) from e

    def _get_context(self, component_id: str) -> ComponentEnergyContext:
        return Context(
            energy_model=self,
            consumer_results=self._consumer_results,
            component_id=component_id,
        )

    def evaluate_energy_usage(self) -> dict[str, EcalcModelResult]:
        energy_components = self.get_energy_components()

        for energy_component in energy_components:
            if hasattr(energy_component, "evaluate_energy_usage"):
                context = self._get_context(energy_component.id)
                self._consumer_results.update(energy_component.evaluate_energy_usage(context=context))

        return self._consumer_results

    def evaluate_emissions(self) -> dict[str, dict[str, EmissionResult]]:
        """
        Calculate emissions for fuel consumers and emitters

        Returns: a mapping from consumer_id to emissions
        """
        for energy_component in self.get_energy_components():
            if isinstance(energy_component, Emitter):
                emission_result = energy_component.evaluate_emissions(
                    energy_context=self._get_context(energy_component.id),
                    energy_model=self,
                )

                if emission_result is not None:
                    self._emission_results[energy_component.id] = emission_result

        return self._emission_results

    def get_graph_result(self) -> GraphResult:
        return GraphResult(
            graph=self._get_graph(),
            consumer_results=self._consumer_results,
            emission_results=self._emission_results,
            variables_map=self.variables,
        )
